"""
Main runners. Bootloads Sniffer class.
"""
from __future__ import print_function, absolute_import
from optparse import OptionParser
from sniffer.scanner import Scanner
from sniffer.runner import Sniffer, ScentSniffer
from sniffer.metadata import __version__
import sys

import colorama
colorama.init()

__all__ = ['run', 'main']

def run(sniffer_instance=None, wait_time=0.5, clear=True, args=(), debug=False):
    """
    Runs the auto tester loop. Internally, the runner instanciates the sniffer_cls and
    scanner class.

    ``sniffer_instance`` The class to run. Usually this is set to but a subclass of scanner.
                    Defaults to Sniffer. Sniffer class documentation for more information.
    ``wait_time``   The time, in seconds, to wait between polls. This is dependent on
                    the underlying scanner implementation. OS-specific libraries may choose
                    to ignore this parameter. Defaults to 0.5 seconds.
    ``clear``       Boolean. Set to True to clear the terminal before running the sniffer,
                    (alias, the unit tests). Defaults to True.
    ``args``        The arguments to pass to the sniffer/test runner. Defaults to ().
    ``debug``       Boolean. Sets the scanner and sniffer in debug mode, printing more internal
                    information. Defaults to False (and should usually be False).
    """
    if sniffer_instance is None:
        sniffer_instance = ScentSniffer()

    if debug:
        scanner = Scanner(sniffer_instance.watch_paths, logger=sys.stdout)
    else:
        scanner = Scanner(sniffer_instance.watch_paths)
    #sniffer = sniffer_cls(tuple(args), clear, debug)
    sniffer_instance.set_up(tuple(args), clear, debug)

    sniffer_instance.observe_scanner(scanner)
    scanner.loop(wait_time)

def main(sniffer_instance=None, test_args=(), progname=sys.argv[0], args=sys.argv[1:]):
    """
    Runs the program. This is used when you want to run this program standalone.

    ``sniffer_instance`` A class (usually subclassed of Sniffer) that hooks into the
                    scanner and handles running the test framework. Defaults to
                    Sniffer instance.
    ``test_args``   This function normally extracts args from ``--test-arg ARG`` command. A
                    preset argument list can be passed. Defaults to an empty tuple.
    ``program``     Program name. Defaults to sys.argv[0].
    ``args``        Command line arguments. Defaults to sys.argv[1:]
    """
    parser = OptionParser(version="%prog " + __version__)
    parser.add_option('-w', '--wait', dest="wait_time", metavar="TIME", default=0.5, type="float",
                      help="Wait time, in seconds, before possibly rerunning tests. "
                      "(default: %default)")
    parser.add_option('--no-clear', dest="clear_on_run", default=True, action="store_false",
                      help="Disable the clearing of screen")
    parser.add_option('--debug', dest="debug", default=False, action="store_true",
                      help="Enabled debugging output. (default: %default)")
    parser.add_option('-x', '--test-arg', dest="test_args", default=[], action="append",
                      help="Arguments to pass to nose (use multiple times to pass multiple "
                      "arguments.)")
    (options, args) = parser.parse_args(args)
    test_args = test_args + tuple(options.test_args)

    if options.debug:
        print("Options:", options)
        print("Test Args:", test_args)
    try:
        print("Starting watch...")
        run(sniffer_instance, options.wait_time, options.clear_on_run, test_args, options.debug)
    except KeyboardInterrupt:
        print("Good bye.")
    except Exception:
        import traceback
        traceback.print_exc()
        return sys.exit(1)
    return sys.exit(0)

if __name__ == '__main__':
    try:
        import psyco
        psyco.full()
    except ImportError:
        pass
    main()



__all__ = [
    '__author__', '__author_email__', '__copyright__', '__credits__',
    '__license__', '__version__'
]


__author__ = "Jeff Hui"
__author_email__ = "jeff@jeffhui.net"
__copyright__ = "Copyright 2013, Jeff Hui"
__credits__ = [
    "Jeff Hui",
    "Patrice Neff",
    "Andrew Lee",
    "Will Harris",
    "Jonas Tingeborn"
]

__license__ = "MIT"
__version__ = "0.3.2"



"""
Scanner that relies on the MacFSEvents (OSX) library.

This is an OS-specific implementation that eliminates the constant polling of the
directory tree by hooking into OSX's IO events.
"""
from __future__ import absolute_import
from .base import BaseScanner
import os
import fsevents
import time
import sys

class FSEventsScanner(BaseScanner):
    """
    This works with MacFSEvents to hook into OSX's file watching mechanisms.
    """
    def __init__(self, *args, **kwargs):
        super(FSEventsScanner, self).__init__(*args, **kwargs)
        self._observer = self._generate_observer()

    def _generate_observer(self):
        observer = fsevents.Observer()
        # use file_events=True to mimic other implementations
        for path in self.paths:
            stream = fsevents.Stream(self._callback, path, file_events=True)
            observer.schedule(stream)
        return observer

    def loop(self, sleep_time=None):
        self.log("Library of choice: MacFSEvents")
        self.trigger_init()
        # using observer.run() doesn't let us catch the keyboard interrupt
        self._observer.start() # separate thread
        try:
            while 1:
                time.sleep(60) # simulate blocking
        except (KeyboardInterrupt, OSError, IOError):
            self.stop()
        #observer.run() # blocking

    def stop(self):
        # Ugly hack, calling Observer.stop() creates a lot of atexit errors
        # that we just want to escape without problems.
        #
        # So we're silently absorbing all the errors
        try:
            from io import StringIO
        except:
            try:
                from cStringIO import StringIO
            except:
                from StringIO import StringIO
        old_err, old_out = sys.stderr, sys.stdout
        sys.stderr, sys.stdout = StringIO(), StringIO()

        self._observer.stop()

        sys.stderr, sys.stdout = old_err, old_out

#    def step(self):
#        observer = self._generate_observer()
#        observer.start()
#        time.sleep(1) # it's really a guess at this point :(
#        observer.stop()
#        observer.join()

    def _callback(self, event):
        if not self.is_valid_type(event.name):
            return
        if event.mask & (fsevents.IN_MODIFY):
            self.trigger_modified(event.name)
        if event.mask & fsevents.IN_CREATE:
            self.trigger_created(event.name)
        if event.mask & fsevents.IN_DELETE:
            self.trigger_deleted(event.name)



from setuptools import setup, find_packages

# will fail to install via pip/easy_install
#from sniffer.metadata import __version__, __author__, __author_email__

__version__, __author__, __author_email__ = "0.3.2", "Jeff Hui", "jeff@jeffhui.net"

setup(
    name='sniffer',
    version=__version__,
    description='An automatic test runner. Supports nose out of the box.',
    long_description=open('README.rst').read(),
    author=__author__,
    author_email=__author_email__,
    url='http://github.com/jeffh/sniffer/',
    install_requires=[
        'colorama',
        'python-termstyle',
        'nose',
    ],
    extras_require = {
        'Growl': ['gntp==0.7'],
        'LibNotify': ['py-notify==0.3.1'],
        'OSX': ['MacFSEvents==0.2.8'],
        #'Windows': ['pywin'], # not part of PYPI
        'Linux': ['pyinotify==0.9.0'],
    },
    entry_points = {
        'console_scripts': ['sniffer = sniffer:main'],
    },
    packages=find_packages(),
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Testing',
        'Topic :: Software Development :: Libraries',
    ],
)



"""
Provides the scanner class which deals with monitoring and notifying changing
in the file system.

It provides a polling technique which is an OS-independent and uses no third-party
libraries at the cost of performance. The polling technique constantly walks through
the directory tree to see which files changed, calling os.stat on the files.
"""
from __future__ import absolute_import
from .base import PollingScanner

__all__ = ['Scanner']

Scanner = PollingScanner

# A messy way to import library-specific classes.
# I should really make a factory class or something, but I'm lazy.
# Plus, factories remind me a lot of java...
def _import(module, cls):
    global Scanner

    try:
        cls = str(cls)
        mod = __import__(str(module), globals(), locals(), [cls], 1)
        Scanner = getattr(mod, cls)
    except ImportError:
        pass

_import('pywin_scanner', 'PyWinScanner') # windows
_import('fsevents_scanner', 'FSEventsScanner') # osx
_import('pyinotify_scanner', 'PyINotifyScanner') # linux



"""
File watching on windows using the Win32API.
Requires the pywin32 library

The code is based off Tim Golden's work:
http://timgolden.me.uk/python/win32_how_do_i/watch_directory_for_changes.html
"""
from __future__ import absolute_import
from .base import BaseScanner
import win32file
import win32con
import os
import thread, Queue, time

ACTIONS = {}
for i, name in enumerate(('Created', 'Deleted', 'Updated', 'Renamed from', 'Renamed to')):
    ACTIONS[i] = name
FILE_LIST_DIR = 0x0001

class PyWinScanner(BaseScanner):
    """
    Scanner built on PyWin32 (aka, Win32API).
    """
    def __init__(self, *args, **kwargs):
        super(PyWinScanner, self).__init__(*args, **kwargs)
        self._running = False
        self._q = Queue.Queue(1)
        
    def _get_handle(self, path):
        return win32file.CreateFile(
            path,                                                 # filename
            FILE_LIST_DIR,                                        # desired access
            win32con.FILE_SHARE_READ | win32con.FILE_SHARE_WRITE, # share mode
            None,                                                 # security attrs
            win32con.OPEN_EXISTING,                               # creation disposition
            win32con.FILE_FLAG_BACKUP_SEMANTICS,                  # flags & attrs
            None                                                  # template file (?)
        )

    def _get_changes_blocking(self, handle):
        result = win32file.ReadDirectoryChangesW(
            handle, # directory handle
            1024, # buffer len (return size)
            True, # watch recursively
            # notify filters
            win32con.FILE_NOTIFY_CHANGE_FILE_NAME | # create/rename/delete files
            #win32con.FILE_NOTIFY_CHANGE_DIR_NAME | # create/rename/delete dirs
            win32con.FILE_NOTIFY_CHANGE_LAST_WRITE, #| # on system write
            #win32con.FILE_NOTIFY_CHANGE_ATTRIBUTES |
            #win32con.FILE_NOTIFY_CHANGE_SIZE |
            #win32con.FILE_NOTIFY_CHANGE_SECURITY,
            None, # overlapped structure (?)
            None  # completion routine (async only)
        )
        self._q.put(result)

    """
      Run the file system monitoring in a separate thread, so that the user
      can interrupt (terminate) the main thread from the console.
    """
    def _get_changes(self, handle):
        thread.start_new_thread(self._get_changes_blocking, (handle,))
        while self._q.empty():
            time.sleep(0.1)
        return self._q.get()

    def step(self):
        for path, handle in ((p, self._get_handle(p)) for p in self.paths):
            results = self._get_changes(handle)
            for action, filename in results:
                fullpath = os.path.join(path, filename)
                if not self.is_valid_type(fullpath):
                    continue
                action = ACTIONS.get(action, "unknown")
                if action == 'Created':
                    self.trigger_created(fullpath)
                elif action in ('Updated', 'Renamed to', 'Renamed from'):
                    self.trigger_modified(fullpath)
                elif action == 'Deleted':
                    self.trigger_deleted(fullpath)
        
    def loop(self, sleep_time=None):
        self.log("Library of choice: PyWin32 (eww)")
        self.trigger_init()
        self._running = True
        while self._running:
            self.step()
            # TODO: figure out if we really need to sleep

    def stop(self):
        self._running = False
